@@ -1,6 +1,86 @@
+//! # Sentinel
+//!
+//! `sentinel` is a sentinel-terminated slice library.
+//!
+//! ## How it works
+//!
+//! ### Rust Slices
+//!
+//! In Rust, the slice type `&[T]` is basically defined like that: `(*const T, usize)`. The `usize`
+//! indicates the number of `T`s referenced at the `*const T`. Knowing in advance the size of an
+//! array, like that, has numerous advantages, which won't be discussed here.
+//!
+//! There is however two main problems with the `&[T]` type:
+//!
+//! 1. It is not (at least, yet) FFI-safe. One cannot create an `extern "C" fn(s: &[u32])` function
+//! and expect it to work when calling it from C-code.
+//!
+//! 2. The size of `&[T]` has the size of two `usize`s.
+//!
+//! ### Sentinels?
+//!
+//! A sentinel is a special value that is used to determine the end of an array. For example, in C,
+//! the `char *` type may be a pointer to a "null-terminated" string. This is an example of
+//! sentinel-terminated slice.
+//!
+//! ```txt
+//! CString:
+//! char *ptr
+//!  |
+//! 'H' 'e' 'l' 'l' 'o' '\0'
+//!                       ^ sentinel, anything after this point may be invalid.
+//! str:
+//! *const u8, 5
+//!  |
+//! 'H' 'e' 'l' 'l' 'o'
+//!                     ^ no sentinel, we know the slice contains 5 elements.
+//! ```
+//!
+//! This crate remains generic over how sentinels are defined. It uses the [`Sentinel`] trait, which
+//! is roughly defined like that:
+//!
+//! ```Rust
+//! trait Sentinel<T> {
+//!     fn is_sentinel(val: &T) -> bool;
+//! }
+//! ```
+//!
+//! It is used to determine whether a specific instance of `T` should be treated as a "sentinel"
+//! value.
+//!
+//! ### SSlice
+//!
+//! Finally, in conjonction with the [`Sentinel`] trait, this crate defines the [`SSlice<T, S>`]
+//! type. It is generic over `T`, the type of stored elements, and over `S: Sentinel<T>`, defining
+//! which instances of `T` should be considered sentinel values.
+//!
+//! ```Rust
+//! struct SSlice<T, S: Sentinel<T>> {
+//!     _marker: PhantomData<(T, S)>,
+//! }
+//! ```
+//!
+//! Note that this type actually contains no data. Only references to this type can be created (i.e.
+//! `&SSlice<T, S>` or `&mut SSlice<T, S>`), and those references have the size a single `usize`.
+//!
+//! ## Features
+//!
+//!  - `alloc` - adds support for the `alloc` crate. This adds the [`SBox<T, S>`] type.
+//!
+//!  - `null` - this feature enables the [`Null`] type, which implements the [`Sentinel`] trait for
+//! common types (integers, pointers, Option<T>).
+//!
+//!  - `nightly` - makes use of the unstable `extern_type` feature to make sure no instance of
+//! [`SSlice<T, S>`] can be created on the stack by making it [`!Sized`]. This feature also enables
+//! support for the new `allocator_api` unstable feature.
+//!
+//! *`null` and `alloc` are enabled by default.*
+
 #![cfg_attr(not(test), no_std)]
 #![cfg_attr(feature = "nightly", feature(extern_types))]
 #![cfg_attr(feature = "nightly", feature(allocator_api))]
+#![cfg_attr(feature = "nightly", feature(core_ffi_c))]
+#![cfg_attr(feature = "nightly", feature(ptr_metadata))]
 
 #[cfg(feature = "alloc")]
 extern crate alloc;
@@ -19,11 +99,6 @@ mod null;
 #[cfg(feature = "null")]
 pub use self::null::*;
 
-#[cfg(feature = "null")]
-mod cstr;
-#[cfg(feature = "null")]
-pub use self::cstr::*;
-
 mod iter;
 pub use self::iter::*;
 
@@ -32,6 +107,8 @@ mod sbox;
 #[cfg(any(feature = "nightly", feature = "alloc"))]
 pub use self::sbox::*;
 
+#[cfg(feature = "null")]
+mod display;
 mod index;
 
 #[cfg(feature = "nightly")]
@@ -64,7 +141,7 @@ impl<T, S: Sentinel<T>> SSlice<T, S> {
     ///
     /// This invalid must remain until the end of the lifetime `'a` (at least).
     #[inline(always)]
-    pub const unsafe fn from_ptr<'a>(ptr: *const T) -> &'a Self {
+    pub unsafe fn from_ptr<'a>(ptr: *const T) -> &'a Self {
         &*(ptr as *const Self)
     }
 
@@ -156,7 +233,7 @@ impl<T, S: Sentinel<T>> SSlice<T, S> {
 
     /// Returns a pointer to the first element that is part of the slice.
     #[inline(always)]
-    pub const fn as_ptr(&self) -> *const T {
+    pub fn as_ptr(&self) -> *const T {
         self as *const Self as *const T
     }
 
@@ -294,19 +371,37 @@ impl<T, S: Sentinel<T>> SSlice<T, S> {
         &mut *self.as_mut_ptr()
     }
 
-    /// Returns a slice referencing every element of this [`SSlice<T, S>`].
+    /// Returns a slice referencing every element of this [`SSlice<T, S>`], not including the
+    /// terminating sentinel character.
     #[inline]
     pub fn as_slice(&self) -> &[T] {
         let len = self.len();
         unsafe { core::slice::from_raw_parts(self.as_ptr(), len) }
     }
 
-    /// Returns a slice referencing every element of this [`SSlice<T, S>`].
+    /// Returns a slice referencing every element of this [`SSlice<T, S>`], not including the
+    /// terminating sentinel character.
     #[inline]
     pub fn as_slice_mut(&mut self) -> &mut [T] {
         let len = self.len();
         unsafe { core::slice::from_raw_parts_mut(self.as_mut_ptr(), len) }
     }
+
+    /// Returns a slice referencing every element of this [`SSlice<T, S>`], including the
+    /// terminating sentinel character.
+    #[inline]
+    pub fn as_slice_with_sentinel(&self) -> &[T] {
+        let len = self.len().wrapping_add(1);
+        unsafe { core::slice::from_raw_parts(self.as_ptr(), len) }
+    }
+
+    /// Returns a slice referencing every element of this [`SSlice<T, S>`], including the
+    /// terminating sentinel character.
+    #[inline]
+    pub fn as_slice_with_sentinel_mut(&mut self) -> &mut [T] {
+        let len = self.len().wrapping_add(1);
+        unsafe { core::slice::from_raw_parts_mut(self.as_mut_ptr(), len) }
+    }
 }
 
 impl<T: Hash, S: Sentinel<T>> Hash for SSlice<T, S> {
